/**
 * Copyright 2013 Peergreen S.A.S.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.shelbie.core.console;

import jline.TerminalFactory;
import jline.console.completer.Completer;
import org.apache.felix.service.command.CommandProcessor;
import org.apache.felix.service.command.CommandSession;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.ow2.shelbie.core.branding.BrandingService;
import org.ow2.shelbie.core.prompt.PromptService;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Collections;

import static org.mockito.Matchers.any;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertTrue;

/**
 * Created with IntelliJ IDEA.
 * User: guillaume
 * Date: 12/02/13
 * Time: 13:43
 * To change this template use File | Settings | File Templates.
 */
public class JLineConsoleTestCase {

    @Mock
    private CommandProcessor processor;
    @Mock
    private CommandSession session;
    @Mock
    private Completer completer;
    @Mock
    private Runnable callback;
    @Mock
    private PromptService prompt;
    @Mock
    private BrandingService branding;

    @BeforeMethod
    public void setUp() throws Exception {
        MockitoAnnotations.initMocks(this);
        when(processor.createSession(any(InputStream.class), any(PrintStream.class), any(PrintStream.class)))
        .thenReturn(session);
    }

    @Test
    public void testExitCallbackIsCalledAndSessionIsClosed() throws Exception {

        ByteArrayInputStream in = new ByteArrayInputStream("".getBytes());
        ByteArrayOutputStream one = new ByteArrayOutputStream();
        PrintStream out = new PrintStream(one);
        ByteArrayOutputStream two = new ByteArrayOutputStream();
        PrintStream err = new PrintStream(two);

        when(session.getConsole()).thenReturn(System.out);

        JLineConsole console = new JLineConsole(processor, completer, in, out, err, TerminalFactory.get());
        console.setCallback(callback);
        console.setPromptService(prompt);

        // Start the console as new Thread
        Thread t = new Thread(console);
        t.start();

        // Sleep some time before closing
        Thread.sleep(50);
        console.close();

        // Wait for the thread to stop
        t.join();

        // Checks
        verify(callback).run();
        verify(session).close();
    }

    @Test
    public void testBrandingServiceUsage() throws Exception {

        ByteArrayInputStream in = new ByteArrayInputStream("".getBytes());
        ByteArrayOutputStream one = new ByteArrayOutputStream();
        PrintStream out = new PrintStream(one);
        ByteArrayOutputStream two = new ByteArrayOutputStream();
        PrintStream err = new PrintStream(two);

        when(branding.getBanner(true)).thenReturn("My Company Brand");
        when(branding.getVariables()).thenReturn(Collections.<String, Object>emptyMap());
        when(session.getConsole()).thenReturn(out);

        JLineConsole console = new JLineConsole(processor, completer, in, out, err, TerminalFactory.get());
        console.setPromptService(prompt);
        console.setBrandingService(branding);

        // Start the console as new Thread
        Thread t = new Thread(console);
        t.start();

        // Sleep some time before closing
        Thread.sleep(50);
        console.close();

        // Wait for the thread to stop
        t.join();

        // Checks
        assertTrue(one.toString().startsWith("My Company Brand"));
    }

    @Test(timeOut = 200)
    public void testIOExceptionSupport() throws Exception {

        ByteArrayInputStream in = new ByteArrayInputStream("do\n".getBytes());
        ByteArrayOutputStream one = new ByteArrayOutputStream();
        PrintStream out = new PrintStream(one);
        ByteArrayOutputStream two = new ByteArrayOutputStream();
        PrintStream err = new PrintStream(two);

        when(session.getConsole()).thenReturn(out);
        when(session.execute(any(CharSequence.class))).thenThrow(new IOException("Simulate broken stream"));

        JLineConsole console = new JLineConsole(processor, completer, in, out, err, TerminalFactory.get());
        console.setPromptService(prompt);

        // Start the console as new Thread
        Thread t = new Thread(console);
        t.start();

        // Wait for the thread to stop by itself
        t.join();
    }
}
